theory Cooper_Code
  imports 
    LinearQuantifierElim.Cooper     
    "HOL-Library.Code_Target_Numeral"
begin

fun nnf :: "atom fm \<Rightarrow> atom fm" where
"nnf (And \<phi>\<^sub>1 \<phi>\<^sub>2) = And (nnf \<phi>\<^sub>1) (nnf \<phi>\<^sub>2)" |
"nnf (Or \<phi>\<^sub>1 \<phi>\<^sub>2) = Or (nnf \<phi>\<^sub>1) (nnf \<phi>\<^sub>2)" |
"nnf (Neg TrueF) = FalseF" |
"nnf (Neg FalseF) = TrueF" |
"nnf (Neg (Neg \<phi>)) = (nnf \<phi>)" |
"nnf (Neg (And \<phi>\<^sub>1 \<phi>\<^sub>2)) = (Or (nnf (Neg \<phi>\<^sub>1)) (nnf (Neg \<phi>\<^sub>2)))" |
"nnf (Neg (Or \<phi>\<^sub>1 \<phi>\<^sub>2)) = (And (nnf (Neg \<phi>\<^sub>1)) (nnf (Neg \<phi>\<^sub>2)))" |
"nnf (Neg (Atom a)) = neg\<^sub>Z a" |
"nnf \<phi> = \<phi>"

fun lift_nnf_qe :: "(atom fm \<Rightarrow> atom fm) \<Rightarrow> atom fm \<Rightarrow> atom fm" where
"lift_nnf_qe qe (And \<phi>\<^sub>1 \<phi>\<^sub>2) = and (lift_nnf_qe qe \<phi>\<^sub>1) (lift_nnf_qe qe \<phi>\<^sub>2)" |
"lift_nnf_qe qe (Or \<phi>\<^sub>1 \<phi>\<^sub>2) = or (lift_nnf_qe qe \<phi>\<^sub>1) (lift_nnf_qe qe \<phi>\<^sub>2)" |
"lift_nnf_qe qe (Neg \<phi>) = neg(lift_nnf_qe qe \<phi>)" |
"lift_nnf_qe qe (ExQ \<phi>) = qe(nnf(lift_nnf_qe qe \<phi>))" |
"lift_nnf_qe qe \<phi> = \<phi>"

lemma [simp]: "Z.nnf f = nnf f" 
  by (induct f rule: nnf.induct, auto)

lemma [simp]: "Z.lift_nnf_qe qe f = lift_nnf_qe qe f" 
  by (induct qe f rule: lift_nnf_qe.induct, auto)

lemma [code]: "qe_cooper = lift_nnf_qe (qe_cooper\<^sub>1 \<circ> hd_coeffs1)"
  unfolding qe_cooper_def
  by (rule fun_cong[of _ _ "(qe_cooper\<^sub>1 \<circ> hd_coeffs1)"], intro ext, simp)

type_synonym var = int
type_synonym constr = "String.literal * (int * var) list * int" 


fun max_var_c :: "constr \<Rightarrow> var" where
  "max_var_c (s, cs, c) = fold max (map snd cs) 0" 

fun constr_to_atom :: "var \<Rightarrow> constr \<Rightarrow> atom" where
  "constr_to_atom n (s, cs, c) = (
    let xs = map (\<lambda> i. case find (\<lambda> p. snd p = i) cs of None \<Rightarrow> 0 | Some p \<Rightarrow> fst p) [0 .. n]
    in if s = STR ''GEQ'' then Le c xs else 
       if s = STR ''LEQ'' then Le (-c) (map uminus xs) else
    Code.abort (STR ''unsupported'') (\<lambda> _. Le 0 []))" 

fun max_var_cs :: "constr list \<Rightarrow> var" where
  "max_var_cs cs = fold max (map max_var_c cs) 0" 

definition constrs_to_fm :: "constr list \<Rightarrow> atom fm" where
  "constrs_to_fm cs = (let n = max_var_cs cs in 
    (ExQ ^^ (nat n + 1)) (list_conj (map (Atom o constr_to_atom n) cs)))" 

fun "interpret" :: "('a \<Rightarrow> 'b list \<Rightarrow> bool) \<Rightarrow> 'a fm \<Rightarrow> 'b list \<Rightarrow> bool" where
"interpret h TrueF xs = True" |
"interpret h FalseF xs = False" |
"interpret h (Atom a) xs = h a xs" |
"interpret h (And \<phi>\<^sub>1 \<phi>\<^sub>2) xs = (interpret h \<phi>\<^sub>1 xs \<and> interpret h \<phi>\<^sub>2 xs)" |
"interpret h (Or \<phi>\<^sub>1 \<phi>\<^sub>2) xs = (interpret h \<phi>\<^sub>1 xs \<or> interpret h \<phi>\<^sub>2 xs)" |
"interpret h (Neg \<phi>) xs = (\<not> interpret h \<phi> xs)" |
"interpret h (ExQ \<phi>) xs = Code.abort (STR ''quantifier'') (\<lambda> _. True)"

lemma qfree_interpret: "qfree f \<Longrightarrow> interpret I\<^sub>Z f = Z.I f"  
  by (induct f, auto)

definition cooper_sat :: "atom fm \<Rightarrow> bool" where
  "cooper_sat f = interpret I\<^sub>Z (qe_cooper f) []" 

lemma cooper_sat: "Z.normal f \<Longrightarrow> cooper_sat f = Z.I f []" 
  unfolding cooper_sat_def
  by (subst qfree_interpret; intro qfree_cooper I_cooper)

definition cooper :: "constr list \<Rightarrow> bool" where
  "cooper = cooper_sat o constrs_to_fm" 

lemma Z_normal_Ex: "Z.normal f \<Longrightarrow> Z.normal ((ExQ ^^ n) f)" 
  by (induct n, auto)

lemma Z_normal_list_conj: "(\<And> f. f \<in> set fs \<Longrightarrow> Z.normal f) \<Longrightarrow> Z.normal (list_conj fs)" 
  unfolding list_conj_def
  by (induct fs, auto)

lemma Z_normal_constrs_to_fm: "Z.normal (constrs_to_fm f)" 
  by (unfold constrs_to_fm_def Let_def)
     (auto intro!: Z_normal_Ex Z_normal_list_conj)

lemma cooper: "cooper f = Z.I (constrs_to_fm f) []" 
  unfolding cooper_def o_def
  by (rule cooper_sat[OF Z_normal_constrs_to_fm])

definition ex_input :: "constr list" where
 "ex_input = [(STR ''GEQ'', [(2,1)], (0)), (STR ''GEQ'', [(-1,0)], 2)]" 

value (code) "qe_cooper (constrs_to_fm ex_input)" 
value (code) "cooper ex_input" 


export_code cooper in Haskell module_name Cooper
end